import base64
import io
import json
import os
import warnings
from collections.abc import Collection, Mapping
from typing import Any, Union

try:
    import geopandas as gpd
    import numpy as np
    import pandas as pd
    import polars as pl
except ImportError:
    raise ImportError(
        "Could not import one of dependencies [geopandas, numpy, pandas, polars].  "
        "Must install nuclei[client] in order to use the utils functions"
    )


DF = "DataFrame"
DF_PARQUET = "DataFrame.parquet"
PDS = "Series"
NUMPY = "numpy"
GDF = "GeoDataFrame"
GPDS = "GeoSeries"
POLARS_DF = "POLARS_DataFrame"
PANDAS_DF = "PANDAS_DataFrame"
PANDAS_DF_PARQUET = "PANDAS_DataFrame.parquet"
PANDAS_PDS = "PANDAS_Series"


def to_json(data: str) -> dict:
    return json.loads(data.replace("'", '"'))


def buffer_to_base64(buf: io.BytesIO) -> str:
    buf.seek(0)
    return base64.b64encode(buf.read()).decode("utf-8")


def base64_to_buffer(blob: bytes) -> io.BytesIO:
    blob = base64.b64decode(blob)
    buf = io.BytesIO()
    buf.write(blob)
    return buf


def deserialize_polars_json(message: dict) -> pl.DataFrame:
    """
    Only needed for backwards compatibility.
    """
    return pl.read_json(io.BytesIO(message["body"].encode("utf8")))


def serialize_polars_json(obj: pl.DataFrame) -> dict:
    return {"nuclei": {"type": POLARS_DF}, "body": obj.write_json()}


def deserialize_pandas_json(message: dict) -> Union[pd.DataFrame, pd.Series]:
    """
    Only needed for backwards compatibility.
    """
    typ = "frame"
    if message["nuclei"]["type"] == PDS or message["nuclei"]["type"] == PANDAS_PDS:
        typ = "series"

    return pd.read_json(message["body"], typ=typ).sort_index()


def serialize_pandas_json(obj: Union[pd.DataFrame, pd.Series]) -> dict:
    if isinstance(obj, pd.DataFrame):
        return {"nuclei": {"type": PANDAS_DF}, "body": obj.to_json()}
    else:
        return {"nuclei": {"type": PANDAS_PDS}, "body": obj.to_json()}


def serialize_pandas_parquet(obj: Union[pd.DataFrame, pd.Series]) -> dict:
    buf = io.BytesIO()
    if isinstance(obj, pd.Series):
        obj = obj.to_frame(obj.name)
    try:
        obj.to_parquet(buf, index=True)
    except Exception:
        return {"nuclei": {"type": PANDAS_DF}, "body": obj.to_json()}

    return {"nuclei": {"type": PANDAS_DF_PARQUET}, "body": buffer_to_base64(buf)}


def deserialize_pandas_parquet(message: dict) -> pd.DataFrame:
    buf = base64_to_buffer(message["body"])
    return pd.read_parquet(buf)


def deserialize_geopandas_json(
    message: dict,
) -> Union[gpd.GeoDataFrame, gpd.GeoSeries, Any]:
    """
    Only needed for backwards compatibility.
    """
    obj = gpd.read_file(message["body"]).reset_index(drop=True)
    obj = obj.drop(columns="id")

    if len(obj.columns) == 1 and obj.columns[0] == "geometry":
        obj = obj.geometry

    return obj


def serialize_geopandas_json(obj: Union[gpd.GeoDataFrame, gpd.GeoSeries]) -> dict:
    if isinstance(obj, gpd.GeoDataFrame):
        return {"nuclei": {"type": GDF}, "body": obj.to_json()}
    else:
        return {"nuclei": {"type": GPDS}, "body": obj.to_json()}


def serialize_numpy(array: np.ndarray) -> dict:
    buf = io.BytesIO()
    np.save(buf, array, allow_pickle=False)
    return {"nuclei": {"type": NUMPY}, "body": buffer_to_base64(buf)}


def deserialize_numpy(message: dict) -> np.ndarray:
    buf = base64_to_buffer(message["body"])
    buf.seek(0)
    return np.load(buf, allow_pickle=False)


def python_types_to_message(
    schema: Union[
        dict,
        pd.Series,
        pd.DataFrame,
        pl.DataFrame,
        gpd.GeoSeries,
        gpd.GeoDataFrame,
        np.ndarray,
        list,
        str,
        float,
        int,
        bool,
    ],
    use_json: bool = True,
) -> Any:
    """
    Cast python types to jsonifyable message.

    Parameters
    ----------
    schema
    use_json
        Added for rollout reasons. Use old json formatting until every backend and the frontend is updated.

    Returns
    -------
    message
        jsonifyable message.
    """
    use_json = not bool(os.environ.get("FORCE_SERIALIZATION", not use_json))
    if gpd is not None:
        if isinstance(schema, (gpd.GeoDataFrame, gpd.GeoSeries)):
            if use_json:
                return serialize_geopandas_json(schema)
            return NotImplemented
    if isinstance(schema, pl.DataFrame):
        if use_json:
            return serialize_polars_json(schema)
        return NotImplemented
    if isinstance(schema, (pd.DataFrame, pd.Series)):
        if use_json:
            return serialize_pandas_json(schema)
        return serialize_pandas_parquet(schema)
    if isinstance(schema, np.ndarray):
        if use_json:
            # no NaN in array
            if all(isinstance(x, (float, int)) for x in schema.flatten()):
                return json.loads(json.dumps(schema.tolist()).replace("NaN", "null"))
            else:
                return schema.tolist()
        return serialize_numpy(schema)
    # recurse and convert to jsonifyable types
    if isinstance(schema, Mapping):
        return {k: python_types_to_message(v) for k, v in schema.items()}
    if isinstance(schema, Collection) and not isinstance(schema, str):
        return [python_types_to_message(k) for k in schema]
    return schema


def message_to_python_types(
    message: Union[
        dict,
        list,
        str,
        float,
        int,
        bool,
    ]
) -> Union[
    dict,
    pd.Series,
    pd.DataFrame,
    pl.DataFrame,
    gpd.GeoSeries,
    gpd.GeoDataFrame,
    np.ndarray,
    list,
    str,
    float,
    int,
    bool,
]:
    """
    Cast json result to python types by searching for objects

    Parameters
    ----------
    message
        parsed json message

    Returns
    -------
    schema
        schema with original python types.
    """

    if isinstance(message, dict):
        # input is a custom serialized type
        if "nuclei" in message and len(message) == 2:
            t = message["nuclei"]["type"]
            if t == DF_PARQUET:
                warnings.warn(
                    "`DataFrame.parquet` is deprecated and will soon be replaced by `PANDAS_DataFrame.parquet`.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return deserialize_pandas_parquet(message)
            elif t == PANDAS_DF_PARQUET:
                return deserialize_pandas_parquet(message)
            elif t == DF or t == PDS:
                warnings.warn(
                    "`DataFrame` and `Series` is deprecated and will soon be replaced by `PANDAS_DataFrame` or "
                    "`PANDAS_Series`.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return deserialize_pandas_json(message)
            elif t == PANDAS_DF or t == PANDAS_PDS:
                return deserialize_pandas_json(message)
            elif t == GDF or t == GPDS:
                return deserialize_geopandas_json(message)
            elif t == POLARS_DF:
                return deserialize_polars_json(message)
            elif t == NUMPY:
                return deserialize_numpy(message)

        # input is a "normal" dict
        return {k: message_to_python_types(v) for k, v in message.items()}

    # input is a list, tuple or set
    if isinstance(message, Collection) and not isinstance(message, str):
        return [message_to_python_types(k) for k in message]

    # return original input if it is anything else
    return message



import logging
import os
from typing import Any

import jwt
import requests


def create_session() -> requests.Session:
    """
    Initialising session object to call the NUCLEI endpoints.
    Provide the user token from https://nuclei.cemsbv.io/#/personal-access-tokens to
    upon request or set as environment variable.

    Returns
    -------
    Session
    """

    # initialising session
    _session = requests.Session()

    def _response_hook(
        r: requests.Response, *args: Any, **kwargs: Any
    ) -> requests.Response:
        """
        Response hook functions to set as event hook on the session.

        This hook does:
            - Update short-lived access token if needed.

        Parameters
        ----------
        r : Response
        args : Any
        kwargs : Any

        Returns
        -------
        Response
        """
        # send new request after update short-lived access token
        try:
            # decode JWT token
            _ = jwt.decode(
                r.request.headers["Authorization"].split(" ")[1],
                algorithms=["HS256"],
                options={"verify_signature": False, "verify_exp": True},
            )
        except jwt.ExpiredSignatureError:
            logging.info(r.text)

            # get short-lived access token
            logging.info("Update Bearer token")
            token = authenticate()

            # set token to header
            _session.headers.update({"Authorization": f"Bearer {token}"})
            r.request.headers["Authorization"] = _session.headers["Authorization"]  # type: ignore

            return _session.send(r.request, verify=False)
        return r

    # set bearer token
    _session.headers.update({"Authorization": f"Bearer {authenticate()}"})

    # set hook
    _session.hooks["response"].append(_response_hook)
    return _session


def authenticate() -> str:
    """
    Returns a validated short-lived token from backend.

    Promts the user for a user-token if it is not stored as an environmental variable
    "NUCLEI_TOKEN". A prompted user token will be stored as env-var after validation.

    Will throw an exception when authentication fails.
    """

    user_token = _get_valid_user_token()

    return _get_valid_shortlived_token(user_token)


def _get_valid_user_token() -> str:
    """
    Obtains the user-token from the env-vars or stdin and returns it.
    If the token is valid, it is stored as env-var "NUCLEI_TOKEN" as a side-effect.
    """

    # check if User Token is in environment variables
    if "NUCLEI_TOKEN" in os.environ:
        logging.info("user token found in environment")
        token = os.environ["NUCLEI_TOKEN"]
        _validate_user_token(token)

    # ask user for user token
    else:
        token = input(
            "Authentication is needed! Please provide your NUCLEI User Token. "
            "You can obtain your NUCLEI User Token on https://nuclei.cemsbv.io/#/personal-access-tokens."
        )

        _validate_user_token(token)
        os.environ["NUCLEI_TOKEN"] = token
        logging.info("user token set in environment")

    return token


def _validate_user_token(token: str) -> None:
    """Validade a JWT User-token by trying to decode it.

    Parameters
    ----------
    token: str
        The Nuclei User JWT.
    """

    try:
        _ = jwt.decode(
            token,
            algorithms=["HS256"],
            options={"verify_signature": False, "verify_exp": True},
        )
    except jwt.ExpiredSignatureError:
        raise jwt.ExpiredSignatureError(
            "Your NUCLEI User Token is expired. Please create a new token on "
            "https://nuclei.cemsbv.io/#/personal-access-tokens."
        )

    except jwt.InvalidTokenError as e:
        raise jwt.InvalidTokenError(
            "Your NUCLEI User Token is invalid. Please copy your user token from "
            "https://nuclei.cemsbv.io/#/personal-access-tokens. "
            f"The following error occurred: {e}"
        )


def _get_valid_shortlived_token(user_token: str) -> str:
    """
    Takes a nuclei user-token and returns a valid nuclei shortlived token.
    Will remove the NUCLEI_TOKEN environmental variable key if no token could be
    obtained with the user-token.
    """

    r = requests.get(
        "https://nuclei.cemsbv.io/v1/shortlived-access-token",
        data=user_token,
    )

    if r.status_code == 400 or r.status_code == 401:
        logging.info("Remove user token from environment")
        _ = os.environ.pop("NUCLEI_TOKEN", None)
        raise Exception("Unauthorized - invalid token")
    elif r.status_code != 200:
        logging.info("Remove user token from environment")
        _ = os.environ.pop("NUCLEI_TOKEN", None)
        raise Exception(
            f"Cannot authorize (status: {r.status_code}, msg: {r.json()['msg']})"
        )

    return r.text



import base64
from functools import lru_cache
from typing import Any, List, Optional, Union

import jwt
import requests

from nuclei import create_session

# try import serialize and deserialize functions
try:
    from IPython.display import Image

    from nuclei.client import utils
except ImportError:
    raise ImportError(
        "Could not import one of dependencies [geopandas, numpy, pandas, polars, ipython].  "
        "You must install nuclei[client] in order to use NucleiClient"
    )

ROUTING = {
    "PileCore": "https://crux-nuclei.com/api/pilecore/v2",
    "VibraCore": "https://crux-nuclei.com/api/vibracore/v1",
    "CPT Core": "https://crux-nuclei.com/api/gef-model",
    "ShallowCore": "https://crux-nuclei.com/api/shallowcore/v1",
    "GEF Map": "https://crux-nuclei.com/api/gef",
}


class NucleiClient:
    def __init__(self) -> None:
        """
        NUCLEI Client.

        This class allows the user to interact with our APIs. API documentation can be
        found on our platform > https://nuclei.cemsbv.io/#
        """
        # initialize session
        self.session = create_session()

        # get routing table to application
        self.routing = ROUTING

    def get_url(self, app: str) -> str:
        """
        Get API's url

        Parameters
        ----------
        app : str
            Application name
        Returns
        -------
        url : str
        """
        if app in self.applications:
            return self.routing[app]
        raise ValueError(
            f"Application not available, please select one of the following valid applications {self.applications}"
        )

    @property
    def user_claims(self) -> List[str]:
        """
        Provide the user claims of your token.

        Returns
        -------
        out : list[str]
            Names of the API's
        """
        return jwt.decode(
            self.session.headers["Authorization"].split(" ")[1],  # type: ignore
            algorithms=["HS256"],
            options={"verify_signature": False, "verify_exp": False},
        )["user_claims"]["allowed_access"].split("|")

    @property
    def applications(self) -> List[str]:
        """
        Provide available API's in the Nuclei landscape.

        Returns
        -------
        out : list[str]
            Names of the API's
        """
        return list(self.routing.keys())

    @lru_cache(16)
    def _get_app_specification(self, app: str) -> dict:
        """
        Private methode to get the JSON schema of the API documentation.

        Parameters
        ----------
        app : str
            Name of the API.
        Returns
        -------
        dict
        """
        return requests.get(self.get_url(app) + "/openapi.json").json()

    def get_endpoints(self, app: str) -> List[str]:
        """
        Get available endpoints of single API.
        Parameters
        ----------
        app : str
            Name of the API.
        Returns
        -------
        out : list[str]
            Endpoint urls.
        """
        return list(self._get_app_specification(app)["paths"].keys())

    def get_endpoint_type(self, app: str, endpoint: str) -> str:
        """
        Parameters
        ----------
        app
            name of the app
        endpoint
            url of the endpoint.
        Returns
        -------
        "get" | "post"
        """

        if endpoint in self.get_endpoints(app):
            return list(self._get_app_specification(app)["paths"][endpoint].keys())[0]
        raise ValueError(
            f"Endpoint name not valid, please select on of the following valid endpoints {self.get_endpoints(app)}"
        )

    def call_endpoint(
        self,
        app: str,
        endpoint: str,
        schema: Optional[Union[dict, str]] = None,
        return_response: bool = False,
    ) -> Any:
        """
        Calls an API in the nuclei landscape.

        Parameters
        ----------
        app: str
            Name of the API. call `get_applications` to obtain a list with all applications.
        endpoint: str
            Name of the API's endpoint. call `get_endpoints` to obtain a list with all applications for a given API.
        schema: dict, optional
            Default is None
            The parameter schema for the API. Take a look at the API documentation.
        return_response: bool, optional
            Default is False
            Return the requests response instead of the parsed data object.

        Returns
        -------
        out : dict
            Parsed API's json response.
        """

        t = self.get_endpoint_type(app, endpoint)

        if isinstance(schema, str):
            schema = utils.to_json(schema)

        if t == "get":
            r = self.session.get(
                self.get_url(app) + endpoint,
                params=utils.python_types_to_message(schema),
            )
        elif t == "post":
            r = self.session.post(
                self.get_url(app) + endpoint,
                json=utils.python_types_to_message(schema),
            )
        else:
            raise ValueError(
                "Not a valid request type. Only GET or POST requests are supported."
            )

        if return_response:
            return r

        content_type = r.headers["Content-Type"]
        if content_type == "image/png;base64":
            return Image(base64.b64decode(r.text))
        elif content_type == "image/png":
            return Image(r.content)
        elif content_type == "application/json":
            result = r.json()
            try:
                python_types = utils.message_to_python_types(result)
            except Exception as e:
                print("Nuclei parsing error", e, result)
                return result
            return python_types
        elif content_type.startswith("text/"):
            return r.text
        return r.content
